''' Main underlying functions for SpatialDE functionality.
'''
import sys
import logging
from time import time
import warnings

import numpy as np
from scipy import optimize
from scipy import linalg
from scipy import stats
from scipy.misc import derivative
from scipy.special import logsumexp

with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    from tqdm.autonotebook import tqdm

import pandas as pd

from .util import qvalue

from anndata import AnnData
import NaiveDE


def get_l_limits(X):
    Xsq = np.sum(np.square(X), 1)
    R2 = -2. * np.dot(X, X.T) + (Xsq[:, None] + Xsq[None, :])
    R2 = np.clip(R2, 0, np.inf)
    R_vals = np.unique(R2.flatten())
    R_vals = R_vals[R_vals > 1e-8]
    
    l_min = np.sqrt(R_vals.min()) / 2.
    l_max = np.sqrt(R_vals.max()) * 2.
    
    return l_min, l_max

## Kernels ##

def SE_kernel(X, l):
    Xsq = np.sum(np.square(X), 1)
    R2 = -2. * np.dot(X, X.T) + (Xsq[:, None] + Xsq[None, :])
    R2 = np.clip(R2, 1e-12, np.inf)
    return np.exp(-R2 / (2 * l ** 2))


def linear_kernel(X):
    K = np.dot(X, X.T)
    return K / K.max()


def cosine_kernel(X, p):
    ''' Periodic kernel as l -> oo in [Lloyd et al 2014]

    Easier interpretable composability with SE?
    '''
    Xsq = np.sum(np.square(X), 1)
    R2 = -2. * np.dot(X, X.T) + (Xsq[:, None] + Xsq[None, :])
    R2 = np.clip(R2, 1e-12, np.inf)
    return np.cos(2 * np.pi * np.sqrt(R2) / p)


def gower_scaling_factor(K):
    ''' Gower normalization factor for covariance matric K

    Based on https://github.com/PMBio/limix/blob/master/limix/utils/preprocess.py
    '''
    n = K.shape[0]
    P = np.eye(n) - np.ones((n, n)) / n
    KP = K - K.mean(0)[:, np.newaxis]
    trPKP = np.sum(P * KP)

    return trPKP / (n - 1)


def factor(K):
    S, U = np.linalg.eigh(K)
    # .clip removes negative eigenvalues
    return U, np.clip(S, 1e-8, None)


def get_UT1(U):
    return U.sum(0)


def get_UTy(U, y):
    return y.dot(U)


def mu_hat(delta, UTy, UT1, S, n, Yvar=None):
    ''' ML Estimate of bias mu, function of delta.
    '''
    if Yvar is None:
        Yvar = np.ones_like(S)

    UT1_scaled = UT1 / (S + delta * Yvar)
    sum_1 = UT1_scaled.dot(UTy)
    sum_2 = UT1_scaled.dot(UT1)

    return sum_1 / sum_2


def s2_t_hat(delta, UTy, S, n, Yvar=None):
    ''' ML Estimate of structured noise, function of delta
    '''
    if Yvar is None:
        Yvar = np.ones_like(S)

    UTy_scaled = UTy / (S + delta * Yvar)
    return UTy_scaled.dot(UTy) / n


def LL(delta, UTy, UT1, S, n, Yvar=None):
    ''' Log-likelihood of GP model as a function of delta.

    The parameter delta is the ratio s2_e / s2_t, where s2_e is the
    observation noise and s2_t is the noise explained by covariance
    in time or space.
    '''

    mu_h = mu_hat(delta, UTy, UT1, S, n, Yvar)
    
    if Yvar is None:
        Yvar = np.ones_like(S)

    sum_1 = (np.square(UTy - UT1 * mu_h) / (S + delta * Yvar)).sum()
    sum_2 = np.log(S + delta * Yvar).sum()

    with np.errstate(divide='ignore'):
        return -0.5 * (n * np.log(2 * np.pi) + n * np.log(sum_1 / n) + sum_2 + n)


def logdelta_prior_lpdf(log_delta):
    s2p = 100.
    return -np.log(np.sqrt(2 * np.pi * s2p)) - np.square(log_delta - 20.) / (2 * s2p)


def make_objective(UTy, UT1, S, n, Yvar=None):
    def LL_obj(log_delta):
        return -LL(np.exp(log_delta), UTy, UT1, S, n, Yvar)

    return LL_obj


def brent_max_LL(UTy, UT1, S, n):
    LL_obj = make_objective(UTy, UT1, S, n)
    o = optimize.minimize_scalar(LL_obj, bounds=[-10, 10], method='bounded', options={'maxiter': 32})
    max_ll = -o.fun
    max_delta = np.exp(o.x)
    max_mu_hat = mu_hat(max_delta, UTy, UT1, S, n)
    max_s2_t_hat = s2_t_hat(max_delta, UTy, S, n)

    return max_ll, max_delta, max_mu_hat, max_s2_t_hat


def lbfgsb_max_LL(UTy, UT1, S, n, Yvar=None):
    LL_obj = make_objective(UTy, UT1, S, n, Yvar)
    min_boundary = -10
    max_boundary = 20.
    x, f, d = optimize.fmin_l_bfgs_b(LL_obj, 0., approx_grad=True,
                                                 bounds=[(min_boundary, max_boundary)],
                                                 maxfun=64, factr=1e12, epsilon=1e-4)
    max_ll = -f
    max_delta = np.exp(x[0])

    boundary_ll = -LL_obj(max_boundary)
    if boundary_ll > max_ll:
        max_ll = boundary_ll
        max_delta = np.exp(max_boundary)

    boundary_ll = -LL_obj(min_boundary)
    if boundary_ll > max_ll:
        max_ll = boundary_ll
        max_delta = np.exp(min_boundary)


    max_mu_hat = mu_hat(max_delta, UTy, UT1, S, n, Yvar)
    max_s2_t_hat = s2_t_hat(max_delta, UTy, S, n, Yvar)

    s2_logdelta = 1. / (derivative(LL_obj, np.log(max_delta), n=2) ** 2)

    return max_ll, max_delta, max_mu_hat, max_s2_t_hat, s2_logdelta


def search_max_LL(UTy, UT1, S, n, num=32):
    ''' Search for delta which maximizes log likelihood.
    '''
    min_obj = np.inf
    max_log_delta = np.nan
    LL_obj = make_objective(UTy, UT1, S, n)
    for log_delta in np.linspace(start=-10, stop=20, num=num):
        cur_obj = LL_obj(log_delta)
        if cur_obj < min_obj:
            min_obj = cur_obj
            max_log_delta = log_delta

    max_delta = np.exp(max_log_delta)
    max_mu_hat = mu_hat(max_delta, UTy, UT1, S, n)
    max_s2_t_hat = s2_t_hat(max_delta, UTy, S, n)
    max_ll = -min_obj

    return max_ll, max_delta, max_mu_hat, max_s2_t_hat


def make_FSV(UTy, S, n, Gower):
    def FSV(log_delta):
        s2_t = s2_t_hat(np.exp(log_delta), UTy, S, n)
        s2_t_g = s2_t * Gower

        return s2_t_g / (s2_t_g + np.exp(log_delta) * s2_t)

    return FSV


def lengthscale_fits(exp_tab, U, UT1, S, Gower, num=64):
    ''' Fit GPs after pre-processing for particular lengthscale
    '''
    results = []
    n, G = exp_tab.shape
    for g in tqdm(range(G), leave=False):
        y = exp_tab.iloc[:, g]
        UTy = get_UTy(U, y)

        t0 = time()
        max_reg_ll, max_delta, max_mu_hat, max_s2_t_hat, s2_logdelta = lbfgsb_max_LL(UTy, UT1, S, n)
        max_ll = max_reg_ll
        t = time() - t0

        # Estimate standard error of Fraction Spatial Variance
        FSV = make_FSV(UTy, S, n, Gower)
        s2_FSV = derivative(FSV, np.log(max_delta), n=1) ** 2 * s2_logdelta
        
        results.append({
            'g': exp_tab.columns[g],
            'max_ll': max_ll,
            'max_delta': max_delta,
            'max_mu_hat': max_mu_hat,
            'max_s2_t_hat': max_s2_t_hat,
            'time': t,
            'n': n,
            'FSV': FSV(np.log(max_delta)),
            's2_FSV': s2_FSV,
            's2_logdelta': s2_logdelta
        })
        
    return pd.DataFrame(results)


def null_fits(exp_tab):
    ''' Get maximum LL for null model
    '''
    results = []
    n, G = exp_tab.shape
    for g in range(G):
        y = exp_tab.iloc[:, g]
        max_mu_hat = 0.
        max_s2_e_hat = np.square(y).sum() / n  # mll estimate
        max_ll = -0.5 * (n * np.log(2 * np.pi) + n + n * np.log(max_s2_e_hat))

        results.append({
            'g': exp_tab.columns[g],
            'max_ll': max_ll,
            'max_delta': np.inf,
            'max_mu_hat': max_mu_hat,
            'max_s2_t_hat': 0.,
            'time': 0,
            'n': n
        })
    
    return pd.DataFrame(results)


def const_fits(exp_tab):
    ''' Get maximum LL for const model
    '''
    results = []
    n, G = exp_tab.shape
    for g in range(G):
        y = exp_tab.iloc[:, g]
        max_mu_hat = y.mean()
        max_s2_e_hat = y.var()
        sum1 = np.square(y - max_mu_hat).sum()
        max_ll = -0.5 * ( n * np.log(max_s2_e_hat) + sum1 / max_s2_e_hat + n * np.log(2 * np.pi) )

        results.append({
            'g': exp_tab.columns[g],
            'max_ll': max_ll,
            'max_delta': np.inf,
            'max_mu_hat': max_mu_hat,
            'max_s2_t_hat': 0.,
            'time': 0,
            'n': n
        })
    
    return pd.DataFrame(results)


def simulate_const_model(MLL_params, N):
    dfm = np.zeros((N, MLL_params.shape[0]))
    for i, params in enumerate(MLL_params.iterrows()):
        params = params[1]
        s2_e = params.max_s2_t_hat * params.max_delta
        dfm[:, i] = np.random.normal(params.max_mu_hat, s2_e, N)
        
    dfm = pd.DataFrame(dfm)
    return dfm


def get_mll_results(results, null_model='const'):
    null_lls = results.query('model == "{}"'.format(null_model))[['g', 'max_ll']]
    model_results = results.query('model != "{}"'.format(null_model))
    model_results = model_results[model_results.groupby(['g'])['max_ll'].transform(max) == model_results['max_ll']]
    mll_results = model_results.merge(null_lls, on='g', suffixes=('', '_null'))
    mll_results['LLR'] = mll_results['max_ll'] - mll_results['max_ll_null']

    return mll_results


def dyn_de(X, exp_tab, kernel_space=None):
    if kernel_space == None:
        kernel_space = {
            'SE': [5., 25., 50.]
        }

    results = []

    if 'null' in kernel_space:
        result = null_fits(exp_tab)
        result['l'] = np.nan
        result['M'] = 1
        result['model'] = 'null'
        results.append(result)

    if 'const' in kernel_space:
        result = const_fits(exp_tab)
        result['l'] = np.nan
        result['M'] = 2
        result['model'] = 'const'
        results.append(result)

    logging.info('Pre-calculating USU^T = K\'s ...')
    US_mats = []
    t0 = time()
    if 'linear' in kernel_space:
        K = linear_kernel(X)
        U, S = factor(K)
        gower = gower_scaling_factor(K)
        UT1 = get_UT1(U)
        US_mats.append({
            'model': 'linear',
            'M': 3,
            'l': np.nan,
            'U': U,
            'S': S,
            'UT1': UT1,
            'Gower': gower
        })

    if 'SE' in kernel_space:
        for lengthscale in kernel_space['SE']:
            K = SE_kernel(X, lengthscale)
            U, S = factor(K)
            gower = gower_scaling_factor(K)
            UT1 = get_UT1(U)
            US_mats.append({
                'model': 'SE',
                'M': 4,
                'l': lengthscale,
                'U': U,
                'S': S,
                'UT1': UT1,
                'Gower': gower
            })

    if 'PER' in kernel_space:
        for period in kernel_space['PER']:
            K = cosine_kernel(X, period)
            U, S = factor(K)
            gower = gower_scaling_factor(K)
            UT1 = get_UT1(U)
            US_mats.append({
                'model': 'PER',
                'M': 4,
                'l': period,
                'U': U,
                'S': S,
                'UT1': UT1,
                'Gower': gower
            })

    t = time() - t0
    logging.info('Done: {0:.2}s'.format(t))

    logging.info('Fitting gene models')
    n_models = len(US_mats)
    for i, cov in enumerate(tqdm(US_mats, desc='Models: ')):
        result = lengthscale_fits(exp_tab, cov['U'], cov['UT1'], cov['S'], cov['Gower'])
        result['l'] = cov['l']
        result['M'] = cov['M']
        result['model'] = cov['model']
        results.append(result)

    n_genes = exp_tab.shape[1]
    logging.info('Finished fitting {} models to {} genes'.format(n_models, n_genes))

    results = pd.concat(results, sort=True).reset_index(drop=True)
    results['BIC'] = -2 * results['max_ll'] + results['M'] * np.log(results['n'])

    return results


def SpatialDE(
    adata: AnnData, 
    use_data: str = "normalized",
    top_genes: int = 10,
    kernel_space = None,
    copy: bool = False,
):
    ''' Perform SpatialDE test

    X : matrix of spatial coordinates times observations
    exp_tab : Expression table, assumed appropriatealy normalised.

    The grid of covariance matrices to search over for the alternative
    model can be specifiec using the kernel_space paramter.
    '''
    sample_info = pd.DataFrame(0, index = adata.obsm[use_data].index, columns = ['imagerow','imagecol','sum_counts'])
    sample_info.loc[:,'imagerow'] = adata.obs['imagerow']
    sample_info.loc[:,'imagecol'] = adata.obs['imagecol']
    sample_info.loc[:,'sum_counts'] = adata.obs['sum_counts']

    norm_expr = NaiveDE.stabilize(adata.obsm[use_data].T).T
    resid_expr = NaiveDE.regress_out(sample_info, norm_expr.T, 'np.log(sum_counts)').T

    sample_resid_expr = resid_expr.sample(n=len(adata.obsm['filtered_counts'].columns), axis=1, random_state=1)
    X = sample_info[['imagerow', 'imagecol']]

    if kernel_space == None:
        l_min, l_max = get_l_limits(X)
        kernel_space = {
            'SE': np.logspace(np.log10(l_min), np.log10(l_max), 10),
            'const': 0
        }

    logging.info('Performing DE test')
    results = dyn_de(X, sample_resid_expr, kernel_space)
    mll_results = get_mll_results(results)

    # Perform significance test
    mll_results['pval'] = 1 - stats.chi2.cdf(mll_results['LLR'], df=1)
    mll_results['qval'] = qvalue(mll_results['pval'])

    adata.uns["SpatialDE_results"] = mll_results

    adata.uns["SpatialDE_top_genes"] = mll_results.sort_values('qval').head(top_genes)['g'].tolist()
    
    print("SpatialDE results stored in adata.uns['SpatialDE_results'] and adata.uns['SpatialDE_top_genes']")

    return adata if copy else None


def model_search(X, exp_tab, DE_mll_results, kernel_space=None):
    ''' Compare model fits with different models.

    This way DE genes can be classified to interpretable function classes.

    The strategy is based on ABCD in the Automatic Statistician, but
    using precomputed covariance matrices for all models in the search space.

    By default searches a grid of periodic covariance matrices and a linear
    covariance matrix.
    '''
    if kernel_space == None:
        P_min, P_max = get_l_limits(X)
        kernel_space = {
            'PER': np.logspace(np.log10(P_min), np.log10(P_max), 10),
            'linear': 0
        }

    de_exp_tab = exp_tab[DE_mll_results['g']]

    logging.info('Performing model search')
    results = dyn_de(X, de_exp_tab, kernel_space)
    new_and_old_results = pd.concat((results, DE_mll_results), sort=True)

    # Calculate model probabilities
    mask = new_and_old_results.groupby(['g', 'model'])['BIC'].transform(min) == new_and_old_results['BIC']
    log_p_data_Hi = -new_and_old_results[mask].pivot_table(values='BIC', index='g', columns='model')
    log_Z = logsumexp(log_p_data_Hi, 1)
    log_p_Hi_data = (log_p_data_Hi.T - log_Z).T
    p_Hi_data = np.exp(log_p_Hi_data).add_suffix('_prob')

    # Select most likely model
    mask = new_and_old_results.groupby('g')['BIC'].transform(min) == new_and_old_results['BIC']
    ms_results = new_and_old_results[mask]

    ms_results = ms_results.join(p_Hi_data, on='g')

    # Retain information from significance testing in the new table
    transfer_columns = ['pval', 'qval', 'max_ll_null']
    ms_results = ms_results.drop(transfer_columns, 1) \
        .merge(DE_mll_results[transfer_columns + ['g']], on='g')

    return ms_results


from typing import Optional, Union
import numpy as np
from anndata import AnnData
from numpy.random.mtrand import RandomState
from scipy.sparse import issparse
import scanpy

def run_diffmap(adata: AnnData, n_comps: int = 15, copy: bool = False):
    """\
    Diffusion Maps [Coifman05]_ [Haghverdi15]_ [Wolf18]_.
    Diffusion maps [Coifman05]_ has been proposed for visualizing single-cell
    data by [Haghverdi15]_. The tool uses the adapted Gaussian kernel suggested
    by [Haghverdi16]_ in the implementation of [Wolf18]_.
    The width ("sigma") of the connectivity kernel is implicitly determined by
    the number of neighbors used to compute the single-cell graph in
    :func:`~scanpy.pp.neighbors`. To reproduce the original implementation
    using a Gaussian kernel, use `method=='gauss'` in
    :func:`~scanpy.pp.neighbors`. To use an exponential kernel, use the default
    `method=='umap'`. Differences between these options shouldn't usually be
    dramatic.
    Parameters
    ----------
    adata
        Annotated data matrix.
    n_comps
        The number of dimensions of the representation.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    `X_diffmap` : :class:`numpy.ndarray` (`adata.obsm`)
        Diffusion map representation of data, which is the right eigen basis of
        the transition matrix with eigenvectors as columns.
    `diffmap_evals` : :class:`numpy.ndarray` (`adata.uns`)
        Array of size (number of eigen vectors).
        Eigenvalues of transition matrix.
    """

    scanpy.tl.diffmap(adata, n_comps=n_comps, copy=copy)

    print(
        "Diffusion Map is done! Generated in adata.obsm['X_diffmap'] nad adata.uns['diffmap_evals']")

    return adata if copy else None


from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np
import networkx as nx
from stlearn._compat import Literal
from typing import Optional, Union
from anndata import AnnData
import warnings
#from .utils import get_img_from_fig, checkType


def cluster_plot(
    adata: AnnData,
    library_id: str = None,
    use_label: str = "louvain",
    list_cluster: list = None,
    data_alpha: float = 1.0,
    cmap: str = "vega_20_scanpy",
    tissue_alpha: float = 1.0,
    title: str = None,
    spot_size: Union[float, int] = 6.5,
    show_color_bar: bool = True,
    show_axis: bool = False,
    show_legend: bool = True,
    dpi: int = 180,
    show_trajectory: bool = False,
    show_subcluster: bool = False,
    name: str = None,
    output: str = None,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Clustering plot for sptial transcriptomics data. Also it has a function to display trajectory inference.

    Parameters
    ----------
    adata
        Annotated data matrix.
    library_id
        Library id stored in AnnData.
    use_label
        Use label result of clustering method.
    list_cluster
        Choose set of clusters that will display in the plot.
    data_alpha
        Opacity of the spot.
    tissue_alpha
        Opacity of the tissue.
    cmap
        Color map to use.
    spot_size
        Size of the spot.
    show_color_bar
        Show color bar or not.
    show_axis
        Show axis or not.
    show_legend
        Show legend or not.
    dpi
        Set dpi as the resolution for the plot.
    show_trajectory
        Show the spatial trajectory or not. It requires stlearn.spatial.trajectory.pseudotimespace.
    show_subcluster
        Show subcluster or not. It requires stlearn.spatial.trajectory.global_level.
    name
        Name of the output figure file.
    output
        Save the figure as file or not.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Nothing
    """

    plt.rcParams['figure.dpi'] = dpi

    n_clusters = len(adata.obs[use_label].unique())

    # Option for turning off showing figure
    plt.ioff()

    # Initialize matplotlib
    fig, a = plt.subplots()

    if show_trajectory:

        if not adata.uns["PTS_graph"]:
            raise ValueError(
                "Please run stlearn.spatial.trajectory.pseudotimespace!")

        tmp = adata.uns["PTS_graph"]

        G = tmp.copy()

        remove = [edge for edge in G.edges if 9999 in edge]
        G.remove_edges_from(remove)
        G.remove_node(9999)
        centroid_dict = adata.uns["centroid_dict"]
        nx.draw_networkx_edges(G, pos=centroid_dict, node_size=1, alpha=1.0,
                               font_size=5, linewidths=1, edge_color='#f4efd3', arrowsize=5, arrowstyle='->', connectionstyle="arc3,rad=0.2")

    from scanpy.plotting import palettes
    if cmap == "vega_10_scanpy":
        cmap = palettes.vega_10_scanpy
    elif cmap == "vega_20_scanpy":
        cmap = palettes.vega_20_scanpy
    elif cmap == "default_102":
        cmap = palettes.default_102
    elif cmap == "default_28":
        cmap = palettes.default_28
    else:
        raise ValueError(
            "We only support vega_10_scanpy, vega_20_scanpy, default_28, default_102")

    cmaps = matplotlib.colors.LinearSegmentedColormap.from_list("", cmap)

    cmap_ = plt.cm.get_cmap(cmaps)

    if list_cluster is not None:
        tmp = adata.obs.loc[adata.obs[use_label].isin(
            np.array(list_cluster).astype(str))]
    else:
        tmp = adata.obs

    # Plot scatter plot based on pixel of spots
    adata.uns["tmp_color"] = []

    for i, cluster in enumerate(tmp.groupby(use_label)):

        _ = a.scatter(cluster[1]['imagecol'], cluster[1]['imagerow'], c=[cmap_(int(i)/19)], label=cluster[0],
                      edgecolor="none", alpha=data_alpha, s=spot_size, marker="o")

        adata.uns["tmp_color"].append(
            matplotlib.colors.to_hex(cmap_(int(i)/19)))

    if show_legend:
        lgnd = a.legend(bbox_to_anchor=(1.2, 1.0), labelspacing=0.05,
                        fontsize=8, handleheight=1., edgecolor='white')
        for handle in lgnd.legendHandles:
            handle.set_sizes([20.0])

    if not show_axis:
        a.axis('off')


    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    image = adata.uns["spatial"][library_id]["images"][adata.uns["spatial"]["use_quality"]]


    # Overlay the tissue image
    a.imshow(image, alpha=tissue_alpha, zorder=-1,)

    if name is None:
        name = use_label

    if output is not None:
        fig.savefig(output + "/" + name + ".png", dpi=dpi,
                    bbox_inches='tight', pad_inches=0)

    if show_subcluster:
        if "sub_cluster_labels" not in adata.obs.columns:
            raise ValueError("Please run stlearn.spatial.cluster.localization")

        for cluster in list_cluster:
            if len(adata.obs[adata.obs[use_label] == str(cluster)]["sub_cluster_labels"].unique()) < 2:
                centroids = [centroidpython(adata.obs[adata.obs[use_label] == str(cluster)][[
                                            "imagecol", "imagerow"]].values)]
                classes = np.array(
                    [adata.obs[adata.obs[use_label] == str(cluster)]["sub_cluster_labels"][0]])

            else:
                from sklearn.neighbors import NearestCentroid
                clf = NearestCentroid()
                clf.fit(adata.obs[adata.obs[use_label] == str(cluster)][["imagecol", "imagerow"]].values,
                        adata.obs[adata.obs[use_label] == str(cluster)]["sub_cluster_labels"])

                centroids = clf.centroids_
                classes = clf.classes_

            for i, label in enumerate(classes):
                if centroids[i][0] < 1500:
                    x = -100
                    y = 50
                else:
                    x = 100
                    y = -50
                a.text(centroids[i][0]+x, centroids[i][1]+y, label, color='black', fontsize=5, zorder=3,
                       bbox=dict(facecolor=adata.uns["tmp_color"][int(cluster)], boxstyle='round', alpha=1.0))

    plt.show()


def centroidpython(data):
    x, y = zip(*data)
    l = len(x)
    return sum(x) / l, sum(y) / l



from .model_zoo import encode, Model
from typing import Optional, Union
from anndata import AnnData
import numpy as np
from .._compat import Literal
from PIL import Image
import pandas as pd
from pathlib import Path

# Test progress bar
from tqdm import tqdm

_CNN_BASE = Literal['resnet50', 'vgg16', 'inception_v3', 'xception']


def extract_feature(
        adata: AnnData,
        cnn_base: _CNN_BASE = 'resnet50',
        n_components: int = 50,
        verbose: bool = False,
        copy: bool = False,
        seeds: int = 1
) -> Optional[AnnData]:
    """\
    Extract latent morphological features from H&E images using pre-trained
    convolutional neural network base

    Parameters
    ----------
    adata
        Annotated data matrix.
    cnn_base
        Established convolutional neural network bases
        choose one from ['resnet50', 'vgg16', 'inception_v3', 'xception']
    n_components
        Number of principal components to compute for latent morphological features
    verbose
        Verbose output
    copy
        Return a copy instead of writing to adata.
    seeds
        Fix random state
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **X_morphology** : `adata.obsm` field
        Dimension reduced latent morphological features.
    """
    feature_df = pd.DataFrame()
    model = Model(cnn_base)

    if "tile_path" not in adata.obs:
        raise ValueError("Please run the function stlearn.pp.tiling")

    with tqdm(total=len(adata), desc="Extract feature", bar_format="{l_bar}{bar} [ time left: {remaining} ]") as pbar:
        for spot, tile_path in adata.obs["tile_path"].items():
            tile = Image.open(tile_path)
            tile = np.asarray(tile, dtype="int32")
            tile = tile.astype(np.float32)
            tile = np.stack([tile])
            if verbose:
                print("extract feature for spot: {}".format(str(spot)))
            features = encode(tile, model)
            feature_df[spot] = features
            pbar.update(1)

    adata.obsm["X_tile_feature"] = feature_df.transpose().to_numpy()

    from sklearn.decomposition import PCA
    pca = PCA(n_components=n_components, random_state=seeds)
    pca.fit(feature_df.transpose().to_numpy())

    adata.obsm["X_morphology"] = pca.transform(
        feature_df.transpose().to_numpy())

    print("The morphology feature is added to adata.obsm['X_morphology']!")

    return adata if copy else None



from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np

from stlearn._compat import Literal
from typing import Optional, Union
from anndata import AnnData
import warnings
#from .utils import get_img_from_fig, checkType
import scanpy

def non_spatial_plot(
    adata: AnnData,
    use_label: str = "louvain",
    dpi: int = 180
) -> Optional[AnnData]:
    
    """\
    A wrap function to plot all the non-spatial plot from scanpy.

    This function will produce 3 plots: PAGA graph, clustering plot in PAGA space and
    DPT in PAGA space.

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_label
        Use label result of clustering method.
    dpi
        Set dpi as the resolution for the plot.
    Returns
    -------
    Nothing
    """

    plt.rcParams['figure.dpi'] = dpi

    if 'paga' in adata.uns.keys():
        adata.uns[use_label+"_colors"] = adata.uns["tmp_color"]

        print("PAGA plot:")

        scanpy.pl.paga(adata, color=use_label)


        scanpy.tl.draw_graph(adata, init_pos='paga')
        adata.uns[use_label+"_colors"] = adata.uns["tmp_color"]


        print("Gene expression (reduced dimension) plot:")
        scanpy.pl.draw_graph(adata, color=use_label, legend_loc='on data')

        print("Diffusion pseudotime plot:")
        scanpy.pl.draw_graph(adata, color="dpt_pseudotime")

    else:

        scanpy.pl.draw_graph(adata)
        adata.uns[use_label+"_colors"] = adata.uns["tmp_color"]

        scanpy.pl.draw_graph(adata, color=use_label, legend_loc='on data')



# Adopted from spatialDE package
from .SpatialDE.base import dyn_de
from .SpatialDE.base import SpatialDE
from .SpatialDE.base import model_search
from .SpatialDE.aeh import fit_patterns
from .SpatialDE.aeh import spatial_patterns
import NaiveDE




from anndata import AnnData
import pandas as pd
from pathlib import Path
import os


def cpdb(
    adata: AnnData,
    cpdb_filepath: str = None,
    sep: str = '\t',
    copy: bool = False,
) -> AnnData:
    """ Add CPDB results into AnnData object

    Parameters
    ----------
    adata: AnnData          The data object to add CPDB results into
    cpdb_filepath: str      The path to the CPDB results file
    sep: str                Separator of the CPDB results file
    copy: bool              Copy flag indicating copy or direct edit

    Returns
    -------
    adata: AnnData          The data object that CPDB results added into

    """
    cpdb = pd.read_csv(cpdb_filepath, sep=sep)
    adata.uns["cpdb"] = cpdb
    print("Added cpdb results to the object!")

    return adata if copy else None



from .tools import clustering
from .tools import enrichment_analysis
from .tools import microenv
from .tools.microenv import cci
from .tools.microenv import SpatialDE
from .tools.microenv import spatial_patterns


from anndata import AnnData
import stlearn
import numpy as np

def PseudoTime(
    adata: AnnData,
    use_label: str ="louvain",
    root_cluster: int = 0,
    root_point: int = 0,
    eps: float = 100,
    radius: int = 50,
    use_sme: bool = False,
    copy: bool = False,

    ) -> AnnData:
	adata.uns["iroot"] = np.flatnonzero(adata.obs[use_label]  == str(root_cluster))[root_point]
	stlearn.spatial.trajectory.pseudotime(data,eps=eps,radius=radius,use_sme = use_sme)

	return adata if copy else None

def PseudoSpaceTime(
    adata: AnnData,
    use_label: str = "louvain",
    local_cluster: int = 0,
    global_cluster: list = [0,1],
    weight: float = 0.5,
    copy: bool = False,

    ) -> AnnData:

	slearn.spatial.trajectory.local_level(data,
		use_label=use_label,
		cluster=local_cluster,
		w=weight)
	slearn.spatial.trajectory.global_level(data,
		use_label=use_label,
		list_cluster=global_cluster,
		w=weight)

	return adata if copy else None


from .embedding.pca import run_pca
from .embedding.umap import run_umap
from .embedding.ica import run_ica
#from .embedding.scvi import run_ldvae
from .embedding.fa import run_fa
from .embedding.diffmap import run_diffmap


import numpy as np
import matplotlib.pyplot as plt
import pandas as pd


def xpercent_scale():
    ''' Helper function to format X-axis to percent tick labels
    '''
    plt.gca().set_xticklabels(['{:.0f}%'.format(x*100) for x in plt.gca().get_xticks()])


def FSV_sig(results, ms_results=None, certain_only=False, covariate_names=['log_total_count']):
    ''' Make a plot of Fraction Spatial Variance vs Q-value

    Optionally provide model selection results to the function will color points by model.

    Point size correspond to certinety of the FSV value.
    '''
    plt.yscale('log')
    
    results = results.copy()
    covariates = results.query('g in @covariate_names')
    results = results.query('g not in @covariate_names').copy()
    
    results['FSV95conf'] = 2 * np.sqrt(results['s2_FSV'])
    if ms_results is not None:
        results = results.merge(ms_results[['g', 'model']],
                                how='outer',
                                on='g',
                                suffixes=('', '_bic'))
    else:
        results['model_bic'] = results['model']
    
    # Split by FSV uncertainty levels
    size_map = {0.0: 40, 0.1: 12, 1.0: 1}
    conf_categories = pd.cut(results['FSV95conf'], [0, 1e-1, 1e0, np.inf])
    for conf_class, result_group in results.groupby(conf_categories):
        if certain_only:
            if conf_class.left > .0:
                continue
        
        # Plot non-signficant genes
        tmp = result_group.query('qval > 0.05')
        label = 'Genes (Not Significant)' if conf_class.left == 0.0 else None
        plt.scatter(tmp['FSV'], tmp['pval'],
                    alpha=0.5,
                    rasterized=True,
                    label=label,
                    marker='o',
                    color='k',
                    s=size_map[conf_class.left])

        tmp = result_group.query('qval <= 0.05')

        # Split significant genes by function class
        model_colors = {'SE': 'C0', 'PER': 'C1', 'linear': 'C2'}
        label_map = {'SE': 'General', 'PER': 'Periodic', 'linear': 'Linear'}
        for model_name, model_group in tmp.groupby('model_bic'):
            label = 'Genes ({} function)'.format(label_map[model_name]) if conf_class.left == 0.0 else None
            plt.scatter(model_group['FSV'], model_group['pval'],
                        alpha=0.5,
                        rasterized=True,
                        label=label,
                        marker='o',
                        color=model_colors[model_name],
                        s=size_map[conf_class.left])

    # Plot external covarites for reference
    plt.scatter(covariates['FSV'], covariates['pval'], marker='x', c='k', s=50, label=None)
    
    FDR_lim = results.query('qval < 0.05')['pval'].max()
    plt.axhline(FDR_lim, ls='--', c='k', lw=1, label='FDR = 0.05')
    
    # Label axes
    plt.xlabel('Fraction spatial variance')
    plt.ylabel('P-value')
    plt.gca().invert_yaxis()

    lgd = plt.legend(scatterpoints=3, loc='upper left', bbox_to_anchor=[1., 1.])
    
    xpercent_scale()


from typing import Optional, Union
from anndata import AnnData
from matplotlib import pyplot as plt
from pathlib import Path
import os
from PIL import Image
Image.MAX_IMAGE_PIXELS = None


def image(
    adata: AnnData,
    imgpath: Union[Path, str],
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Adding image data to the Anndata object

    Parameters
    ----------
    adata
        Annotated data matrix.
    imgpath
        Image path.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **tissue_img** : `adata.uns` field
        Array format of image, saving by Pillow package.
    """
    

    if imgpath is not None and os.path.isfile(imgpath):
        try:
            img = plt.imread(imgpath, 0)
            adata.uns["spatial"] = {}
            adata.uns["spatial"]["Slide-seq"] = {}
            adata.uns["spatial"]["Slide-seq"]["images"] = {}
            adata.uns["spatial"]["Slide-seq"]["images"]["hires"] = img
            adata.uns["spatial"]["use_quality"] = "hires"
            adata.uns["spatial"]["Slide-seq"]["scalefactors"] = {}
            adata.uns["spatial"]["Slide-seq"]["scalefactors"]["tissue_hires_scalef"] = 1

            print("Added tissue image to the object!")

            return adata if copy else None
        except:
            raise ValueError(f'''\
            {imgpath!r} does not end on a valid extension.
            ''')
    else:
        raise ValueError(f'''\
        {imgpath!r} does not end on a valid extension.
        ''')
    return adata if copy else None



from typing import Optional
import numpy as np
from anndata import AnnData
from ..._compat import Literal
import scipy.spatial as spatial
from tqdm import tqdm

_SIMILARITY_MATRIX = Literal["cosine", "euclidean", "pearson", "spearman"]


def adjust(
        adata: AnnData,
        use_data: str = "X_pca",
        radius: float = 50.0,
        rates: int = 1,
        method="mean",
        copy: bool = False,
        similarity_matrix: _SIMILARITY_MATRIX = "cosine"

) -> Optional[AnnData]:
    """\
    SME normalisation: Using spot location information and tissue morphological
    features to correct spot gene expression

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_data
        Input date to be adjusted by morphological features.
        choose one from ["raw", "X_pca", "X_umap"]
    radius
        Radius to select neighbour spots.
    rates
        Strength for adjustment.
    method
        Method for disk smoothing.
        choose one from ["means", "median"]
    copy
        Return a copy instead of writing to adata.
    similarity_matrix
        Matrix to calculate morphological similarity of two spots
        choose one from ["cosine", "euclidean", "pearson", "spearman"]
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **[use_data]_morphology** : `adata.obsm` field
        Add SME normalised gene expression matrix
    """
    if "X_morphology" not in adata.obsm:
        raise ValueError("Please run the function stlearn.pp.extract_feature")
    coor = adata.obs[["imagecol", "imagerow"]]
    if use_data == "raw":
        count_embed = adata.X
    else:
        count_embed = adata.obsm[use_data]
    point_tree = spatial.cKDTree(coor)
    img_embed = adata.obsm["X_morphology"]
    lag_coor = []
    with tqdm(total=len(adata), desc="Adjusting data",
              bar_format="{l_bar}{bar} [ time left: {remaining} ]") as pbar:
        for i in range(len(coor)):
            current_neightbor = point_tree.query_ball_point(
                coor.values[i], radius)  # Spatial weight
            current_neightbor.remove(i)
            if len(current_neightbor) > 0:
                main_count = count_embed[i].reshape(1, -1)
                main_img = img_embed[i].reshape(1, -1)
                surrounding_count = count_embed[current_neightbor]
                surrounding_img = img_embed[current_neightbor]

                similarity = []

                for i in surrounding_img:
                    i = i.reshape(1, -1)  # reshape feature to (1, n_feature)

                    if similarity_matrix == "cosine":
                        from sklearn.metrics.pairwise import cosine_similarity
                        cosine = cosine_similarity(main_img, i)[0][0]
                        cosine = (abs(cosine) + cosine) / 2
                        similarity.append(cosine)
                    elif similarity_matrix == "euclidean":
                        from sklearn.metrics.pairwise import euclidean_distances
                        eculidean = euclidean_distances(main_img, i)[0][0]
                        eculidean = 1 / (1 + eculidean)
                        similarity.append(eculidean)
                    elif similarity_matrix == "pearson":
                        from scipy.stats import pearsonr
                        pearson_corr = abs(
                            pearsonr(main_img.reshape(-1), i.reshape(-1))[0])
                        similarity.append(pearson_corr)
                    elif similarity_matrix == "spearman":
                        from scipy.stats import spearmanr
                        spearmanr_corr = abs(
                            spearmanr(main_img.reshape(-1), i.reshape(-1))[0])
                        similarity.append(spearmanr_corr)

                similarity = np.array(similarity).reshape((-1, 1))
                surrounding_count_adjusted = np.multiply(
                    surrounding_count, similarity)

                for i in range(0, rates):
                    if method == "median":
                        main_count = np.append(main_count,
                                               np.median(surrounding_count_adjusted, axis=0).reshape(1, -1),
                                               axis=0)
                    elif method == "mean":
                        main_count = np.append(main_count,
                                               np.mean(surrounding_count_adjusted, axis=0).reshape(1, -1),
                                               axis=0)
                    elif method == "sum":
                        main_count = np.append(main_count,
                                               np.sum(surrounding_count_adjusted, axis=0).reshape(1, -1),
                                               axis=0)
                    else:
                        raise ValueError("Only 'median' and 'mean' are aceptable")
                lag_coor.append(list(np.sum(main_count, axis=0)))
                pbar.update(1)
            else:
                lag_coor.append(list(np.sum(main_count, axis=0)))
                pbar.update(1)
    key_added = use_data + "_morphology"
    adata.obsm[key_added] = np.array(lag_coor)

    print("The data adjusted by morphology is added to adata.obsm['" + key_added + "']")

    return adata if copy else None



from .adds.add_image import image
from .adds.add_positions import positions
from .adds.parsing import parsing
from .adds.add_lr import lr
from .adds.annotation import annotation
from .adds.add_cpdb import cpdb


import logging as logg
from typing import Union, Optional, Tuple, Collection, Sequence, Iterable
from anndata import AnnData
import numpy as np
from scipy.sparse import issparse, isspmatrix_csr, csr_matrix, spmatrix
from numpy.random.mtrand import RandomState
import scanpy


def run_pca(
    data: Union[AnnData, np.ndarray, spmatrix],
    n_comps: int = 50,
    zero_center: Optional[bool] = True,
    svd_solver: str = 'auto',
    random_state: Optional[Union[int, RandomState]] = 0,
    return_info: bool = False,
    use_highly_variable: Optional[bool] = None,
    dtype: str = 'float32',
    copy: bool = False,
    chunked: bool = False,
    chunk_size: Optional[int] = None,
) -> Union[AnnData, np.ndarray, spmatrix]:
    """\
    Wrap function scanpy.pp.pca
    Principal component analysis [Pedregosa11]_.
    Computes PCA coordinates, loadings and variance decomposition.
    Uses the implementation of *scikit-learn* [Pedregosa11]_.
    Parameters
    ----------
    data
        The (annotated) data matrix of shape `n_obs` × `n_vars`.
        Rows correspond to cells and columns to genes.
    n_comps
        Number of principal components to compute.
    zero_center
        If `True`, compute standard PCA from covariance matrix.
        If `False`, omit zero-centering variables
        (uses :class:`~sklearn.decomposition.TruncatedSVD`),
        which allows to handle sparse input efficiently.
        Passing `None` decides automatically based on sparseness of the data.
    svd_solver
        SVD solver to use:
        `'arpack'`
          for the ARPACK wrapper in SciPy (:func:`~scipy.sparse.linalg.svds`)
        `'randomized'`
          for the randomized algorithm due to Halko (2009).
        `'auto'` (the default)
          chooses automatically depending on the size of the problem.
    random_state
        Change to use different initial states for the optimization.
    return_info
        Only relevant when not passing an :class:`~anndata.AnnData`:
        see “**Returns**”.
    use_highly_variable
        Whether to use highly variable genes only, stored in
        `.var['highly_variable']`.
        By default uses them if they have been determined beforehand.
    dtype
        Numpy data type string to which to convert the result.
    copy
        If an :class:`~anndata.AnnData` is passed, determines whether a copy
        is returned. Is ignored otherwise.
    chunked
        If `True`, perform an incremental PCA on segments of `chunk_size`.
        The incremental PCA automatically zero centers and ignores settings of
        `random_seed` and `svd_solver`. If `False`, perform a full PCA.
    chunk_size
        Number of observations to include in each chunk.
        Required if `chunked=True` was passed.
    Returns
    -------
    X_pca : :class:`~scipy.sparse.spmatrix`, :class:`~numpy.ndarray`
        If `data` is array-like and `return_info=False` was passed,
        this function only returns `X_pca`…
    adata : anndata.AnnData
        …otherwise if `copy=True` it returns or else adds fields to `adata`:
        `.obsm['X_pca']`
             PCA representation of data.
        `.varm['PCs']`
             The principal components containing the loadings.
        `.uns['pca']['variance_ratio']`
             Ratio of explained variance.
        `.uns['pca']['variance']`
             Explained variance, equivalent to the eigenvalues of the
             covariance matrix.
    """

    scanpy.pp.pca(data, n_comps=n_comps, zero_center=zero_center, svd_solver=svd_solver,
            random_state=random_state, return_info=return_info,
            use_highly_variable=use_highly_variable,
            dtype=dtype, copy=copy, chunked=chunked, chunk_size=chunk_size)

    print("PCA is done! Generated in adata.obsm['X_pca'], adata.uns['pca'] and adata.varm['PCs']")


from typing import Optional, Union
from anndata import AnnData
from pathlib import Path
import os
import pandas as pd


def lr(
    adata: AnnData,
    copy: bool = False,
) -> Optional[AnnData]:
    """ Add significant Ligand-Receptor pairs into AnnData object

    Parameters
    ----------
    adata: AnnData          The data object to add L-R info into
    copy: bool              Copy flag indicating copy or direct edit

    Returns
    -------
    adata: AnnData          The data object that L-R added into

    """
    lr = adata.uns['cpdb']['interacting_pair'].to_list()
    lr2 = [i for i in lr if 'complex' not in i]
    lr3 = [i for i in lr2 if ' ' not in i]
    lr4 = [i for i in lr3 if i.count('_') == 1]
    adata.uns['lr'] = lr4
    print("Added ligand receptor pairs to adata.uns['lr'].")

    return adata if copy else None



from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np

from stlearn._compat import Literal
from typing import Optional, Union
from anndata import AnnData
import warnings
#from .utils import get_img_from_fig, checkType


def gene_plot(
    adata: AnnData,
    method: str = "CumSum",
    genes: Optional[Union[str, list]] = None,
    library_id: str = None,
    data_alpha: float = 1.0,
    tissue_alpha: float = 1.0,
    cmap: str = "Spectral_r",
    spot_size: Union[float, int] = 6.5,
    show_legend: bool = False,
    show_color_bar: bool = True,
    show_axis: bool = False,
    dpi: int = 192,
    name: str = None,
    output: str = None,
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Gene expression plot for sptial transcriptomics data.

    Parameters
    ----------
    adata
        Annotated data matrix.
    library_id
        Library id stored in AnnData.
    method
        Use method to count. We prorive: NaiveMean, NaiveSum, CumSum.
    genes
        Choose a gene or a list of genes.
    data_alpha
        Opacity of the spot.
    tissue_alpha
        Opacity of the tissue.
    cmap
        Color map to use.
    spot_size
        Size of the spot.
    show_color_bar
        Show color bar or not.
    show_axis
        Show axis or not.
    show_legend
        Show legend or not.
    dpi
        Set dpi as the resolution for the plot.
    show_trajectory
        Show the spatial trajectory or not. It requires stlearn.spatial.trajectory.pseudotimespace.
    show_subcluster
        Show subcluster or not. It requires stlearn.spatial.trajectory.global_level.
    name
        Name of the output figure file.
    output
        Save the figure as file or not.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Nothing
    """
    plt.rcParams['figure.dpi'] = dpi

    if type(genes) == str:
        genes = [genes]
    colors = _gene_plot(adata, method, genes)

    # Option for turning off showing figure
    plt.ioff()

    # Initialize matplotlib
    fig, a = plt.subplots()

    vmin = min(colors)
    vmax = max(colors)
    # Plot scatter plot based on pixel of spots
    plot = a.scatter(adata.obs["imagecol"], adata.obs["imagerow"], edgecolor="none", alpha=data_alpha, s=spot_size, marker="o",
                     vmin=vmin, vmax=vmax, cmap=plt.get_cmap(cmap), c=colors)

    if show_color_bar:

        cb = plt.colorbar(plot, cax=fig.add_axes(
            [0.78, 0.3, 0.03, 0.38]), cmap=cmap)
        cb.outline.set_visible(False)

    if not show_axis:
        a.axis('off')

    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    image = adata.uns["spatial"][library_id]["images"][adata.uns["spatial"]["use_quality"]]
    # Overlay the tissue image
    a.imshow(image, alpha=tissue_alpha, zorder=-1,)

    if name is None:
        name = method
    if output is not None:
        fig.savefig(output + "/" + name + ".png", dpi=dpi,
                    bbox_inches='tight', pad_inches=0)

    plt.show()
    # Store figure
    #fig_np = get_img_from_fig(fig,dpi)
    # plt.close(fig)
    # if "plots" not in adata.uns:
    #    adata.uns['plots'] = {}
    #adata.uns['plots'].update({str(', '.join(genes)):fig_np})
    # print("The plot stored in adata.uns['plot']['" +
    #    str(', '.join(genes)) + "']")


def _gene_plot(adata, method, genes):

    # Gene plot option

    if len(genes) == 0:
        raise ValueError('Genes shoule be provided, please input genes')

    elif len(genes) == 1:

        if genes[0] not in adata.var.index:
            raise ValueError(
                genes[0] + ' is not exist in the data, please try another gene')

        colors = list(adata[:, genes].X.toarray().reshape(1, -1)[0])

        return colors
    else:

        for gene in genes:
            if gene not in adata.var.index:
                genes.remove(gene)
                warnings.warn("We removed " + gene +
                              " because they not exist in the data")

            if len(genes) == 0:
                raise ValueError(
                    'All provided genes are not exist in the data')

        count_gene = adata[:, genes].to_df()

        if method is None:
            raise ValueError(
                'Please provide method to combine genes by NaiveMean/NaiveSum/CumSum')

        if method == "NaiveMean":
            present_genes = (count_gene > 0).sum(axis=1) / len(genes)

            count_gene = (count_gene.mean(axis=1)) * present_genes
        elif method == "NaiveSum":
            present_genes = (count_gene > 0).sum(axis=1) / len(genes)

            count_gene = (count_gene.sum(axis=1)) * present_genes

        elif method == "CumSum":
            count_gene = count_gene.cumsum(axis=1).iloc[:, -1]

        colors = count_gene

        return colors



from matplotlib import pyplot as plt
import matplotlib
import pandas as pd
import numpy as np
import seaborn as sns
import sys
from anndata import AnnData
from typing import Optional, Union

def het_plot(
    adata: AnnData,
    het: pd.DataFrame,
    use_cluster: str = 'louvain',
    dpi: int = 100,
    spot_size: Union[float,int] = 6.5,
    quality: str = "hires",
    name: str = None,
    output: str = None,
):
    """ Plot tissue clusters and cluster heterogeneity using heatmap
    Parameters
    ----------
    adata: AnnData                  The data object to plot
    het: pd.DataFrame               Cluster heterogeneity count results from tl.cci.het
    use_cluster: str                The clustering results to use
    dpi: bool                       Dots per inch
    spot_size: Union[float,int]     Spot size
    
    Returns
    -------
    N/A
    """

    

    plt.rcParams['figure.dpi'] = dpi
    fig, ax = plt.subplots()
    num_clusters = len(set(adata.obs[use_cluster])) + 1
    ax.set_prop_cycle('color',plt.cm.rainbow(np.linspace(0,1,num_clusters)))
    for item in set(adata.obs[use_cluster]):
        ax.scatter(np.array(adata.obs[adata.obs[use_cluster]==item]['imagecol']), 
                   -np.array(adata.obs[adata.obs[use_cluster]==item]['imagerow']), 
                   alpha=0.6, s=spot_size, edgecolors='none')
    ax.legend(range(num_clusters))
    ax.grid(False)
    plt.axis('equal')

    if name is None:
        name = use_cluster
    if output is not None:
        fig.savefig(output + "/" + name + "_scatter.pdf", dpi=dpi, bbox_inches='tight', pad_inches=0)

    plt.rcParams['figure.dpi'] = dpi * 0.8
    plt.subplots()
    sns.heatmap(het)
    plt.axis('equal')

    if output is not None:
        plt.savefig(output + "/" + name + "_heatmap.pdf", dpi=dpi, bbox_inches='tight', pad_inches=0)


def violin_plot(
    adata: AnnData,
    lr: str,
    use_cluster: str = 'louvain',
    dpi: int = 100,
    name: str = None,
    output: str = None,
):
    """ Plot the distribution of CCI counts within spots of each CCI clusters
    Parameters
    ----------
    adata: AnnData          The data object to plot
    lr: str                 The specified Ligand-Receptor pair to plot
    use_cluster: str        The clustering results to use
    dpi: bool               Dots per inch
    name: str               Save as file name
    output: str             Save to directory
    Returns
    -------
    N/A
    """
    try:
        violin = adata.obsm['lr_neighbours'][[lr]]
    except:
        sys.exit('Please run cci counting and clustering first.')
    violin.columns = ['LR_counts']
    violin['cci_cluster'] = adata.obs['lr_neighbours_' + use_cluster]
    plt.rcParams['figure.dpi'] = dpi
    sns.violinplot(x='cci_cluster', y='LR_counts', data=violin, orient='v')
    if name is None:
        name = use_cluster

    if output is not None:
        plt.savefig(output + "/" + name + ".pdf", dpi=dpi, bbox_inches='tight', pad_inches=0)

def stacked_bar_plot(
    adata: AnnData,
    use_annotation: str,
    dpi: int = 100,
    name: str = None,
    output: str = None,
):
    """ Plot the proportion of cell types in each CCI cluster
    Parameters
    ----------
    adata: AnnData          The data object to plot
    use_annotation: str     The cell type annotation to be used in plotting
    dpi: bool               Dots per inch
    name: str               Save as file name
    output: str             Save to directory
    Returns
    -------
    N/A
    """
    sns.set()
    try:
        cci = adata.obs['lr_neighbours_louvain']
    except:
        sys.exit('Please run cci counting and clustering first.')
    try:
        label = adata.obs[use_annotation]
    except:
        sys.exit('spot cell type not found in data.obs[' + use_annotation + ']')
    df = pd.DataFrame(0, index=sorted(set(cci)), columns=set(label))
    for spot in cci.index:
        df.loc[cci[spot], label[spot]] += 1

    # From raw value to percentage
    df2 = df.div(df.sum(axis=1), axis=0)
    plt.rcParams['figure.dpi'] = dpi
    df2.plot(kind='bar', stacked='True', legend=False)
    plt.legend(loc='upper right', bbox_to_anchor=(1.5, 1), ncol=1)
    if name is None:
        name = use_annotation

    if output is not None:
        plt.savefig(output + "/" + name + ".pdf", dpi=dpi, bbox_inches='tight', pad_inches=0)


from typing import Optional, Union
import numpy as np
from anndata import AnnData
import logging as logg
import scipy.spatial as spatial


def disk(
    adata: AnnData,
    use_data: str = "X_umap",
    radius: float = 10.0,
    rates: int = 1,
    method: str = "mean",
    copy: bool = False,
) -> Optional[AnnData]:

    coor = adata.obs[["imagecol", "imagerow"]]
    count_embed = adata.obsm[use_data]
    point_tree = spatial.cKDTree(coor)

    lag_coor = []
    tmp = []

    for i in range(len(coor)):
        current_neightbor = point_tree.query_ball_point(
            coor.values[i], radius)  # Spatial weight
        # print(coor.values[i])
        tmp.append(current_neightbor)
        # print(coor.values[current_neightbor])
        main = count_embed[current_neightbor]
        current_neightbor.remove(i)
        addition = count_embed[current_neightbor]

        for i in range(0, rates):
            main = np.append(main, addition, axis=0)
        if method == "mean":
            # New umap based on SW
            lag_coor.append(list(np.mean(main, axis=0)))
        elif method == "median":
            # New umap based on SW
            lag_coor.append(list(np.median(main, axis=0)))
        else:
            raise ValueError("Only 'median' and 'mean' are aceptable")

    new_embed = use_data + "_disk"

    adata.obsm[new_embed] = np.array(lag_coor)

    print(
        'Disk smoothing function is applied! The new data are stored in adata.obsm["X_diffmap_disk"]')

    return adata if copy else None



from anndata import AnnData
from typing import Optional, Union
import numpy as np
import pandas as pd
import networkx as nx
from scipy.spatial.distance import cdist
import scanpy

def pseudotime(
    adata: AnnData,
    use_label: str = "louvain",
    eps: float = 20,
    threshold: float = 0.01,
    radius: int = 50,
    method: str = "mean",
    use_sme: bool = False,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Perform pseudotime analysis.

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_label
        Use label result of clustering method.
    eps
        The maximum distance between two samples for one to be considered as 
        in the neighborhood of the other. This is not a maximum bound on the 
        distances of points within a cluster. This is the most important DBSCAN 
        parameter to choose appropriately for your data set and distance function.
    threshold
        Threshold to find the significant connection for PAGA graph.
    radius
        radius to adjust data for diffusion map
    method
        method to adjust the data.
    use_sme
        Use adjusted feature by SME normalization or not
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Anndata
    """

    # Localize
    from stlearn.spatials.clustering import localization
    localization(adata, use_label=use_label, eps=eps)

    # Running paga
    scanpy.tl.paga(adata, groups=use_label)

    # Denoising the graph
    scanpy.tl.diffmap(adata)

    if use_sme:
        from stlearn.spatials.morphology import adjust
        adjust(adata, use_data="X_diffmap", radius=radius, method=method)
        adata.obsm["X_diffmap"] = adata.obsm["X_diffmap_morphology"]


    # Get connection matrix
    cnt_matrix = adata.uns["paga"]["connectivities"].toarray()

    # Filter by threshold
    threshold = threshold

    cnt_matrix[cnt_matrix < threshold] = 0.
    cnt_matrix = pd.DataFrame(cnt_matrix)

    # Mapping louvain label to subcluster
    split_node = {}
    for label in adata.obs[use_label].unique():
        split_node[int(label)] = list(
            adata.obs[adata.obs[use_label] == label]["sub_cluster_labels"].unique())

    adata.uns["split_node"] = split_node
    # Replicate louvain label row to prepare for subcluster connection 
    # matrix construction
    replicate_list = np.array([])
    for i in range(0, len(cnt_matrix)):
        replicate_list = np.concatenate(
            [replicate_list, np.array([i]*len(split_node[i]))])

    # Connection matrix for subcluster
    cnt_matrix = cnt_matrix.loc[replicate_list.astype(
        int), replicate_list.astype(int)]

    # Replace column and index
    cnt_matrix.columns = replace_with_dict(cnt_matrix.columns, split_node)
    cnt_matrix.index = replace_with_dict(cnt_matrix.index, split_node)

    # Sort column and index
    cnt_matrix = cnt_matrix.loc[selection_sort(np.array(cnt_matrix.columns)),
                                selection_sort(np.array(cnt_matrix.index))]

    # Create a connection graph of subclusters
    G = nx.from_numpy_matrix(cnt_matrix.values)

    adata.uns['global_graph'] = G

    # Create centroid dict for subclusters
    from sklearn.neighbors import NearestCentroid
    clf = NearestCentroid()
    clf.fit(adata.obs[["imagecol", "imagerow"]].values,
            adata.obs["sub_cluster_labels"])
    centroid_dict = dict(zip(clf.classes_.astype(int), clf.centroids_))
    adata.uns["centroid_dict"] = centroid_dict

    # Running diffusion pseudo-time
    scanpy.tl.dpt(adata)

    return adata if copy else None


######## utils ########

def replace_with_dict(ar, dic):
    # Extract out keys and values
    k = np.array(list(dic.keys()))
    v = np.array(list(dic.values()))

    out = np.zeros_like(ar)
    for key, val in zip(k, v):
        out[ar == key] = val
    return out


def selection_sort(x):
    for i in range(len(x)):
        swap = i + np.argmin(x[i:])
        (x[i], x[swap]) = (x[swap], x[i])
    return x
